﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Net;
using System.Configuration;
using System.Threading;
using SIPSorcery.SIP;
using SIPSorcery.SIP.App;
using SIPSorcery.Net;
using log4net;

namespace SIPTest
{
    class Program
    {
        static IPEndPoint mLocal, mRemote;
        static int mTimes;

        static void Main(string[] args)
        {
            GetSetting();

            int n = 1;
            if (args.Length > 0)
            {
                int.TryParse(args[0], out n);
            }

            mTimes = n;

            StartTest();

            GC.Collect();
            Console.WriteLine("GC Collected @ " + DateTime.Now.ToString());

            Console.ReadLine();
        }

        private static void StartTest()
        {
            Program p = new Program();
            p.Prepare();

            string cmd;

            while (!(cmd = Console.ReadLine()).Equals("END"))
            {
                switch (cmd)
                {
                    case "gc":
                        GC.Collect();
                        Console.WriteLine("GC Collected @ " + DateTime.Now.ToString());
                        break;

                    case "r":
                        Thread t = new Thread(RunTest);
                        t.Start(p);
                        break;
                }
            }

            p.Stop();
        }


        static void RunTest(object state)
        {
            Program p = state as Program;

            string c1 = "1890010";
            string c2 = "1331000";

            for (int i = 0; i < mTimes; i++)
            {
                int n = (int)(DateTime.Now.Ticks % 10000);
                string num = n.ToString("0000");

                p.Call(c1 + num, c2 + num, mRemote);

                Thread.Sleep(20000);
            }
        }

        static void GetSetting()
        {
            string s;
            string[] ss;

            s = ConfigurationManager.AppSettings["Local"];

            if (!string.IsNullOrEmpty(s))
            {
                ss = s.Split(':');
                try
                {
                    mLocal = new System.Net.IPEndPoint(System.Net.IPAddress.Parse(ss[0]), int.Parse(ss[1]));
                }
                catch
                { }
            }

            s = ConfigurationManager.AppSettings["Remote"];

            if (!string.IsNullOrEmpty(s))
            {
                ss = s.Split(':');
                try
                {
                    mRemote = new System.Net.IPEndPoint(System.Net.IPAddress.Parse(ss[0]), int.Parse(ss[1]));
                }
                catch
                { }
            }
        }

        public void Prepare()
        {
            m_sipEp = mLocal;
            m_localIP = m_sipEp.ToString();
            rnd = new Random(DateTime.Now.GetHashCode());
            Start();
        }

        #region Member fields
        private SIPTransport m_sipTransport;

        protected IPEndPoint m_sipEp;

        protected string m_localIP;

        private bool started;

        public bool IsReady;

        Random rnd;
        #endregion

        void Start()
        {
            if (!started)
            {
                started = true;
                Thread t = new Thread(InitialiseSIP);
                t.Name = "InitialiseSIP";
                //t.IsBackground = true;
                t.Start();
            }
        }

        public void Stop()
        {
            IsReady = false;

            if (started)
            {
                started = false;

                if (m_sipTransport != null)
                {
                    m_sipTransport.SIPTransportRequestReceived -= SIPTransportRequestReceived;

                    m_sipTransport.Shutdown();

                    m_sipTransport = null;
                }
            }
        }

        private void InitialiseSIP()
        {
            // Configure the SIP transport layer.
            m_sipTransport = new SIPTransport(SIPDNSManager.ResolveSIPService, new SIPTransactionEngine());
            m_sipTransport.AddSIPChannel(new SIPUDPChannel(m_sipEp));

            m_sipTransport.SIPTransportRequestReceived += SIPTransportRequestReceived;

            IsReady = true;
        }

        private void SIPTransportRequestReceived(SIPEndPoint localSIPEndPoint, SIPEndPoint remoteEndPoint, SIPRequest sipRequest)
        {
            if (sipRequest.Method == SIPMethodsEnum.ACK) return;

            string key = sipRequest.Header.CallId;

            Console.WriteLine("Request: {0}, Key: {1}", sipRequest.Method, key);

            SipRequestInfo sri = new SipRequestInfo
            {
                Key = key,
                Request = sipRequest,
                LocalPoint = localSIPEndPoint,
                RemotePoint = remoteEndPoint
            };

            Thread t = new Thread(ProcessRequest);
            t.Name = "ProcessRequest";
            t.Start(sri);
        }

        private void ProcessRequest(object state)
        {
            try
            {
                SipRequestInfo sri = state as SipRequestInfo;

                if (sri == null) return;

                string key = sri.Key;
                SIPRequest sipRequest = sri.Request;
                SIPEndPoint remoteEndPoint = sri.RemotePoint;
                SIPEndPoint localSIPEndPoint = sri.LocalPoint;

                switch (sipRequest.Method)
                {
                    case SIPMethodsEnum.BYE:
                        OnHangup(key, sipRequest, remoteEndPoint);
                        break;

                    case SIPMethodsEnum.INVITE:
                        OnInvite(key, sipRequest, localSIPEndPoint, remoteEndPoint);
                        break;

                    case SIPMethodsEnum.CANCEL:
                        OnCancel(key, sipRequest, localSIPEndPoint, remoteEndPoint);
                        break;

                    default:
                        OnOther(key, sipRequest, remoteEndPoint);
                        break;
                }
            }
            catch (Exception ex)
            {
                WriteException("ProcessRequest", ex);
            }
        }

        protected void OnHangup(string key, SIPRequest sipRequest, SIPEndPoint remoteEndPoint)
        {
            Channel ch = GetChannel(key);

            lock (colCalls)
            {
                colCalls.Remove(key);
            }

            if (ch != null)
            {
                ch.OnHangup();

                if (ch.callIn)
                {
                    SIPServerUserAgent uas = GetInCallUserAgent(key);

                    if (uas != null)
                    {
                        lock (colUAS)
                        {
                            colUAS.Remove(key);
                        }
                    }

                    if (uas != null && uas.SIPDialogue != null && key == uas.SIPDialogue.CallId)
                    {
                        // Call has been hungup by remote end.
                        AppendTraceMessage("Call hungup by client: " + m_sipEp + "<-" + remoteEndPoint + " " + sipRequest.URI.ToString() + ".\r\n");

                        SIPResponse res = SIPTransport.GetResponse(sipRequest, SIPResponseStatusCodesEnum.Ok, null);

                        //REMARK: debug used only
                        res.Body = " BYE from UAS";

                        Transport.SendResponse(res);

                        if (colProcessedCalls.ContainsKey(key))
                        {
                            colProcessedCalls[key]++;
                        }
                        else
                        {
                            colProcessedCalls[key] = 1;
                        }
                    }

                    uas.CallCancelled -= UASCallCancelled;
                    uas = null;
                }
                else
                {
                    SIPClientUserAgent uac = GetOutCallUserAgent(key);

                    if (uac != null)
                    {
                        lock (colUAC)
                        {
                            colUAC.Remove(key);
                        }
                    }

                    if (uac != null && uac.SIPDialogue != null && key == uac.SIPDialogue.CallId)
                    {
                        // Call has been hungup by remote end.
                        AppendTraceMessage("Call hungup by server: " + m_sipEp + "<-" + remoteEndPoint + " " + sipRequest.URI.ToString() + ".\r\n");

                        SIPResponse res = SIPTransport.GetResponse(sipRequest, SIPResponseStatusCodesEnum.Ok, null);

                        //REMARK: debug used only
                        res.Body = " BYE from UAC";

                        Transport.SendResponse(res);

                        colProcessedCalls[key] = 1;
                    }

                    uac.CallTrying -= CallTrying;
                    uac.CallRinging -= CallRinging;
                    uac.CallAnswered -= CallAnswered;
                    uac.CallFailed -= CallFailed;

                    uac = null;
                }
            }
            else
            {
                if (colProcessedCalls.ContainsKey(key) && colProcessedCalls[key] > 0)
                {
                    // Resend items, just ignore them since already processed
                }
                else
                {
                    AppendTraceMessage("Unmatched BYE request received for " + sipRequest.URI.ToString() + ".\r\n");
                    // 481 error code
                    SIPResponse noCallLegResponse = SIPTransport.GetResponse(sipRequest, SIPResponseStatusCodesEnum.CallLegTransactionDoesNotExist, null);
                    Transport.SendResponse(noCallLegResponse);
                }
            }
        }

        protected void OnCancel(string key, SIPRequest sipRequest, SIPEndPoint localSIPEndPoint, SIPEndPoint remoteEndPoint)
        {
            UASInviteTransaction inviteTransaction = (UASInviteTransaction)Transport.GetTransaction(SIPTransaction.GetRequestTransactionId(key, SIPMethodsEnum.INVITE));

            lock (colUAS)
            {
                colUAS.Remove(key);
            }

            if (inviteTransaction != null)
            {
                AppendTraceMessage("Matching CANCEL request received " + sipRequest.URI.ToString() + ".\r\n");

                SIPCancelTransaction cancelTransaction = Transport.CreateCancelTransaction(sipRequest, remoteEndPoint, localSIPEndPoint, inviteTransaction);
                cancelTransaction.GotRequest(localSIPEndPoint, remoteEndPoint, sipRequest);
            }
            else
            {
                AppendTraceMessage("No matching transaction was found for CANCEL to " + sipRequest.URI.ToString() + ".\r\n");
                SIPResponse noCallLegResponse = SIPTransport.GetResponse(sipRequest, SIPResponseStatusCodesEnum.CallLegTransactionDoesNotExist, null);
                Transport.SendResponse(noCallLegResponse);
            }
        }

        protected void OnInvite(string key, SIPRequest sipRequest, SIPEndPoint localSIPEndPoint, SIPEndPoint remoteEndPoint)
        {
            AppendTraceMessage("Incoming call request: " + localSIPEndPoint + "<-" + remoteEndPoint + " " + sipRequest.URI.ToString() + ".\r\n");

            // Send Trying when get INVITE to prevent dup coming INVITE
            SIPResponse res = SIPTransport.GetResponse(sipRequest, SIPResponseStatusCodesEnum.Trying, null);
            Transport.SendResponse(res);
            Thread.Sleep(200);

            res = SIPTransport.GetResponse(sipRequest, SIPResponseStatusCodesEnum.Ringing, null);
            Transport.SendResponse(res);
            Thread.Sleep(10);

            if (colUAS.ContainsKey(key))
            {
                // Resend items, just ignore them since already processed
            }
            else
            {
                if (colProcessingCalls.ContainsKey(key))
                {
                    // Resend items, just ignore them since already processed
                }
                else
                {
                    colProcessingCalls[key] = 1;
                    OnComingCall(localSIPEndPoint, remoteEndPoint, sipRequest, key);
                }
            }
        }

        protected void OnOther(string key, SIPRequest sipRequest, SIPEndPoint remoteEndPoint)
        {
            AppendTraceMessage("SIP " + sipRequest.Method + " request received but no processing has been set up for it, return OK here.\r\n");
            SIPResponse res = SIPTransport.GetResponse(sipRequest, SIPResponseStatusCodesEnum.Ok, null);
            res.Header.To.ToTag = CallProperties.CreateNewTag();
            Transport.SendResponse(res);
        }

        private void OnComingCall(SIPEndPoint localSIPEndPoint, SIPEndPoint remoteEndPoint, SIPRequest sipRequest, string key)
        {
            UASInviteTransaction uasTransaction = Transport.CreateUASTransaction(sipRequest, remoteEndPoint, localSIPEndPoint, null);
            SIPServerUserAgent uas = new SIPServerUserAgent(Transport, null, null, null, SIPCallDirection.In, null, null, LogTraceMessage, uasTransaction);
            uas.CallCancelled += UASCallCancelled;
            lock (colUAS)
            {
                colUAS[key] = uas;
            }
            Channel ch = null;
            try
            {
                ch = SIPHelper.CreateChannel(sipRequest, remoteEndPoint);
                ch.CallID = key;
                ch.Controller = this;

                lock (colCalls)
                {
                    colCalls[key] = ch;
                }

                string toTag;

                ch.ToTag = toTag = CallProperties.CreateNewTag();

                string sdp = SIPHelper.CreateSDP(mLocal.Address.ToString()).ToString();
                Thread.Sleep(300);
                
                uas.Answer("application/sdp", sdp, toTag, null, SIPDialogueTransferModesEnum.NotAllowed);

                Thread.Sleep(15000);
                uas.SIPDialogue.Hangup(m_sipTransport, null);
            }
            catch (Exception ex)
            {
                uas.Reject(SIPResponseStatusCodesEnum.BusyHere, null, null);
                WriteException("OnComingCall", ex);
            }
        }

        private void UASCallCancelled(ISIPServerUserAgent uas)
        {
            AppendTraceMessage("Incoming call cancelled for: " + uas.CallDestination + "\r\n");

            Channel ch = null;

            string key = uas.CallRequest.Header.CallId;

            ch = GetChannel(key);

            if (ch != null)
            {
                ch.OnCanceled();
            }
        }

        #region Trace Related items
        public void WriteException(string place, Exception ex)
        {
            Console.WriteLine("Exception in : {0} \r\n{1}\r\n", place, ex.ToString());

            if (logger.IsErrorEnabled)
            {
                logger.Error(place, ex);
            }
        }

        public void AppendTraceMessage(string message)
        {
            Console.WriteLine(message);

            if (logger.IsInfoEnabled)
            {
                logger.Info(message);
            }
        }

        /// <summary>
        /// 记录SIP消息日志，使用<SIP></SIP>标识
        /// </summary>
        /// <param name="message">SIP消息内容</param>
        protected void AppendSIPTransportTraceMessage(string message)
        {
            message = "<SIP>\r\n" + message + "\r\n</SIP>";

            Console.WriteLine(message);

            if (logger.IsDebugEnabled)
            {
                logger.Debug(message);
            }
        }

        protected void LogTraceMessage(SIPMonitorEvent monitorEvent)
        {
            if (monitorEvent is SIPMonitorConsoleEvent)
            {
                if (((SIPMonitorConsoleEvent)monitorEvent).EventType != SIPMonitorEventTypesEnum.FullSIPTrace &&
                    ((SIPMonitorConsoleEvent)monitorEvent).EventType != SIPMonitorEventTypesEnum.SIPTransaction)
                {
                    AppendTraceMessage(monitorEvent.Message + "\r\n");
                }
            }
        }
        #endregion

        #region Properties
        protected SIPTransport Transport
        {
            get { return m_sipTransport; }
        }

        ILog logger = SIPSorcery.Sys.AppState.logger;

        protected ILog Logger
        {
            get { return logger; }
        }
        #endregion

        Dictionary<string, SIPClientUserAgent> colUAC = new Dictionary<string, SIPClientUserAgent>();

        Dictionary<string, SIPServerUserAgent> colUAS = new Dictionary<string, SIPServerUserAgent>();

        Dictionary<string, int> colProcessedCalls = new Dictionary<string, int>();

        Dictionary<string, int> colProcessingCalls = new Dictionary<string, int>();

        Dictionary<string, Channel> colCalls = new Dictionary<string, Channel>();

        private void PlaceCall(object state)
        {
            if (state == null) return;

            Channel info = state as Channel;

            if (info != null)
            {
                AppendTraceMessage("Starting call to " + info.Callee + ".\r\n");

                SIPEndPoint proxyEP = new SIPEndPoint(mRemote);
                SIPClientUserAgent uac = new SIPClientUserAgent(m_sipTransport, proxyEP, null, null, LogTraceMessage);
                uac.CallTrying += CallTrying;
                uac.CallRinging += CallRinging;
                uac.CallAnswered += CallAnswered;
                uac.CallFailed += CallFailed;

                SIPCallDescriptor callDescriptor = SIPHelper.GetCallDescriptor(info, m_sipEp);
                uac.Call(callDescriptor);
                string key;

                try
                {
                    if (uac.ServerTransaction == null)
                        throw new NullReferenceException();
                    key = uac.ServerTransaction.TransactionRequest.Header.CallId;
                }
                catch (Exception ex)
                {
                    WriteException("PlaceCall", ex);
                    key = callDescriptor.GetHashCode().ToString();
                }

                info.CallID = key;
                info.Signal.Set();

                lock (colUAC)
                {
                    colUAC[key] = uac;
                }

                lock (colCalls)
                {
                    colCalls[key] = info;
                }
            }
        }

        private void CallRinging(ISIPClientUserAgent uac, SIPResponse sipResponse)
        {
            AppendTraceMessage("Call ringing: " + sipResponse.StatusCode + " " + sipResponse.ReasonPhrase + ".\r\n");
        }

        private void CallTrying(ISIPClientUserAgent uac, SIPResponse sipResponse)
        {
            AppendTraceMessage("Call trying: " + sipResponse.StatusCode + " " + sipResponse.ReasonPhrase + ".\r\n");
        }

        private void CallFailed(ISIPClientUserAgent uac, string errorMessage)
        {
            AppendTraceMessage("Call failed: " + errorMessage + "\r\n");
        }

        private void CallAnswered(ISIPClientUserAgent uac, SIPResponse sipResponse)
        {
            AppendTraceMessage("Call answered: " + sipResponse.StatusCode + " " + sipResponse.ReasonPhrase + ".\r\n");
        }

        public Channel Call(string caller, string callee, EndPoint remotePoint)
        {
            Channel ch = CreateChannel(caller, callee, (IPEndPoint)remotePoint);

            if (ch == null) return null;

            ch.Controller = this;

            Thread t = new Thread(PlaceCall);
            t.Name = "CC:PlaceCall";
            t.Start(ch);
            if (ch.Signal.WaitOne(5000))
            {
                return ch;
            }
            else
            {
                return null;
            }
        }

        Channel CreateChannel(string caller, string callee, IPEndPoint ep)
        {
            Channel ret = new Channel();

            ret.Caller = caller;
            ret.Callee = callee;
            ret.Controller = this;
            ret.RemotePoint = ep;
            ret.callIn = false;

            return ret;
        }

        SIPServerUserAgent GetInCallUserAgent(string key)
        {
            SIPServerUserAgent uas = null;

            try
            {
                lock (colUAS)
                {
                    if (colUAS.ContainsKey(key))
                    {
                        uas = colUAS[key];
                    }
                }
            }
            catch (Exception ex)
            {
                WriteException("GetInCallUserAgent", ex);
            }

            return uas;
        }

        SIPClientUserAgent GetOutCallUserAgent(string key)
        {
            SIPClientUserAgent uac = null;

            try
            {
                lock (colUAC)
                {
                    if (colUAC.ContainsKey(key))
                    {
                        uac = colUAC[key];
                    }
                }
            }
            catch (Exception ex)
            {
                WriteException("GetOutCallUserAgent", ex);
            }

            return uac;
        }

        Channel GetChannel(string key)
        {
            Channel ch = null;

            try
            {
                lock (colCalls)
                {
                    if (colCalls.ContainsKey(key))
                    {
                        ch = colCalls[key];
                    }
                }
            }
            catch (Exception ex)
            {
                WriteException("GetChannel", ex);
            }

            return ch;
        }

        class SipRequestInfo
        {
            public SIPRequest Request { get; set; }
            public string Key { get; set; }
            public SIPEndPoint LocalPoint { get; set; }
            public SIPEndPoint RemotePoint { get; set; }
        }

        public class Channel
        {
            public bool callIn;
            public string CallID;
            public Program Controller;
            public string ToTag;
            public string Callee;
            public string Caller;
            public IPEndPoint RemotePoint;
            internal ManualResetEvent Signal = new ManualResetEvent(false);

            public void OnHangup()
            {
                Console.WriteLine("Hanguped: IN = {2} ID = {0}, Num = {1}", CallID, Caller, callIn);
            }

            public void OnCanceled()
            { }
        }

        class SIPHelper
        {
            public static Channel CreateChannel(SIPRequest req, SIPEndPoint ep)
            {
                Channel ret = null;

                if (req != null)
                {
                    try
                    {
                        SDP reqSdp;
                        SIPViaHeader via;

                        if ((via = req.Header.Vias.TopViaHeader) != null &&
                            (reqSdp = SDP.ParseSDPDescription(req.Body)) != null)
                        {
                            ret = new Channel
                            {
                                Caller = req.Header.From.FromURI.User,
                                Callee = req.Header.To.ToURI.User,
                                RemotePoint = ep.GetIPEndPoint(),
                                callIn = true,
                            };
                        }
                    }
                    catch
                    {
                        throw;
                    }
                }

                return ret;
            }

            public static SDP CreateSDP(string ip)
            {
                SDP sdp = new SDP();

                Random rnd = new Random(DateTime.Now.GetHashCode());
                sdp.SessionId = rnd.Next(1, 10).ToString();
                sdp.AnnouncementVersion = rnd.Next(1283000000, 1284000000);
                sdp.Address = ip;
                sdp.AddMedia(sdp.Address, rnd.Next(40000, 42000));
                sdp.Media[0].FormatList = "8";
                sdp.SessionName = "SIP Phone";
                sdp.Timing = "0 0";
                sdp.AddExtra("a=rtpmap:8 PCMA/8000");

                return sdp;
            }

            public static string MakeSipUrl(string number, string ip, bool withDisplayName)
            {
                string s;

                if (withDisplayName)
                {
                    s = string.Format("{0}<sip:{0}@{1}>", number, ip);
                }
                else
                {
                    s = string.Format("sip:{0}@{1}", number, ip);
                }

                return s;
            }

            public static string MakeSipUrl(string number, string ip)
            {
                return MakeSipUrl(number, ip, false);
            }

            internal static SIPCallDescriptor GetCallDescriptor(Channel info, IPEndPoint localEP)
            {
                if (info != null)
                {
                    string toIP = ((IPEndPoint)info.RemotePoint).Address.ToString();
                    string ip = localEP.Address.ToString();

                    SIPCallDescriptor ret = new SIPCallDescriptor("phone", null, MakeSipUrl(info.Callee, toIP),
                        MakeSipUrl(info.Caller, ip, true), MakeSipUrl(info.Callee, toIP, true),
                        null, null, "phone", SIPCallDirection.Out, null, null, null);

                    ret.ContactUri = SIPURI.ParseSIPURI(string.Format("sip:{0}@{1}", info.Caller, localEP.ToString()));

                    SDP sdp = CreateSDP(mLocal.Address.ToString());

                    ret.Content = sdp.ToString();
                    ret.ContentType = "application/sdp";

                    return ret;
                }

                return null;
            }
        }

        enum CallAction
        {
            Accept,
            Connect,
            Reject,
            Cancel,
        }
    }
}
